import imaplib
import itertools
import logging
import poplib

from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.utils.encoding import force_bytes, force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.metadata.models import MetadataType
from mayan.apps.storage.models import SharedUploadedFile

from ..classes import SourceBackend
from ..exceptions import SourceException

from .literals import (
    DEFAULT_EMAIL_IMAP_MAILBOX, DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
    DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
    DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME, DEFAULT_EMAIL_POP3_TIMEOUT,
    SOURCE_INTERVAL_UNCOMPRESS_CHOICES
)
from .mixins import (
    SourceBackendCompressedMixin, SourceBackendPeriodicMixin,
    SourceBaseMixin
)

__all__ = (
    'SourceBackendEmailMixin', 'SourceBackendIMAPEmail',
    'SourceBackendPOP3Email'
)
logger = logging.getLogger(name=__name__)


class SourceBackendEmailMixin:
    @classmethod
    def get_setup_form_schema(cls):
        result = super().get_setup_form_schema()

        result['fields'].update(
            {
                'host': {
                    'class': 'django.forms.CharField',
                    'label': _('Host'),
                    'kwargs': {
                        'max_length': 128
                    },
                    'required': True
                },
                'ssl': {
                    'class': 'django.forms.BooleanField',
                    'default': True,
                    'label': _('SSL'),
                    'required': False
                },
                'port': {
                    'class': 'django.forms.IntegerField',
                    'help_text': _(
                        'Typical choices are 110 for POP3, 995 for POP3 '
                        'over SSL, 143 for IMAP, 993 for IMAP over SSL.'
                    ),
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _('Port'),
                },
                'username': {
                    'class': 'django.forms.CharField',
                    'kargs': {
                        'max_length': 128,
                    },
                    'label': _('Username'),
                },
                'password': {
                    'class': 'django.forms.CharField',
                    'kargs': {
                        'max_length': 128,
                    },
                    'label': _('Password'),
                    'required': False,
                },
                'metadata_attachment_name': {
                    'class': 'django.forms.CharField',
                    'default': DEFAULT_EMAIL_METADATA_ATTACHMENT_NAME,
                    'help_text': _(
                        'Name of the attachment that will contains the metadata type '
                        'names and value pairs to be assigned to the rest of the '
                        'downloaded attachments.'
                    ),
                    'kargs': {
                        'max_length': 128,
                    },
                    'label': _('Metadata attachment name'),
                },
                'from_metadata_type_id': {
                    'blank': True,
                    'class': 'django.forms.ChoiceField',
                    'help_text': _(
                        'Select a metadata type to store the email\'s '
                        '"from" value. Must be a valid metadata type for '
                        'the document type selected previously.'
                    ),
                    'kwargs': {
                        'choices': itertools.chain(
                            [(None, '---------')],
                            [(instance.id, instance) for instance in MetadataType.objects.all()],
                        )
                    },
                    'label': _('From metadata type'),
                    'null': True,
                    'required': False
                },
                'subject_metadata_type_id': {
                    'blank': True,
                    'class': 'django.forms.ChoiceField',
                    'help_text': _(
                        'Select a metadata type to store the email\'s '
                        'subject value. Must be a valid metadata type for '
                        'the document type selected previously.'
                    ),
                    'kwargs': {
                        'choices': itertools.chain(
                            [(None, '---------')],
                            [(instance.id, instance) for instance in MetadataType.objects.all()],
                        )
                    },
                    'label': _('Subject metadata type'),
                    'null': True,
                    'required': False
                },
                'message_id_metadata_type_id': {
                    'blank': True,
                    'class': 'django.forms.ChoiceField',
                    'help_text': _(
                        'Select a metadata type to store the email\'s '
                        'message ID value. Must be a valid metadata type '
                        'for the document type selected previously.'
                    ),
                    'kwargs': {
                        'choices': itertools.chain(
                            [(None, '---------')],
                            [(instance.id, instance) for instance in MetadataType.objects.all()],
                        )
                    },
                    'label': _('Message ID metadata type'),
                    'null': True,
                    'required': False
                },
                'store_body': {
                    'class': 'django.forms.BooleanField',
                    'default': True,
                    'help_text': _(
                        'Store the body of the email as a text document.'
                    ),
                    'label': _('Store email body'),
                    'required': False
                }
            }
        )
        result['field_order'] = (
            'host', 'ssl', 'port', 'username', 'password',
            'metadata_attachment_name', 'from_metadata_type_id',
            'subject_metadata_type_id', 'message_id_metadata_type_id',
            'store_body'
        ) + result['field_order']

        result['widgets'].update(
            {
                'password': {
                    'class': 'django.forms.widgets.PasswordInput', 'kwargs': {
                        'render_value': True
                    }
                },
                'from_metadata_type_id': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'},
                    }
                },
                'subject_metadata_type_id': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'},
                    }
                },
                'message_id_metadata_type_id': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'},
                    }
                }
            }
        )

        return result

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.document_metadata = {}

    def process_message(self, message):
        from flanker import mime

        message = mime.from_string(string=force_bytes(s=message))

        shared_uploaded_files = self._process_message(message=message)

        # Process source metadata after messages to avoid the metadata
        # attachment to be used to override the source metadata.

        from_metadata_type = self.get_from_metadata_type()
        if from_metadata_type:
            self.document_metadata[from_metadata_type.pk] = message.headers.get('From')

        subject_metadata_type = self.get_subject_metadata_type()
        if subject_metadata_type:
            self.document_metadata[subject_metadata_type.pk] = message.headers.get('Subject')

        message_id_metadata_type = self.get_message_id_metadata_type()
        if message_id_metadata_type:
            self.document_metadata[message_id_metadata_type.pk] = message.headers.get('Message-ID')

        return shared_uploaded_files

    def _process_message(self, message):
        counter = 1
        shared_uploaded_files = []

        # Messages are tree based, do nested processing of message parts until
        # a message with no children is found, then work out way up.
        if message.parts:
            for part in message.parts:
                part_shared_uploaded_files = self._process_message(
                    message=part
                )

                shared_uploaded_files.extend(part_shared_uploaded_files)
        else:
            # Treat inlines as attachments, both are extracted and saved as
            # documents.
            if message.is_attachment() or message.is_inline():
                # Reject zero length attachments.
                if len(message.body) == 0:
                    return shared_uploaded_files

                label = message.detected_file_name or 'attachment-{}'.format(counter)
                counter = counter + 1

                with ContentFile(content=message.body, name=label) as file_object:
                    if label == self.kwargs['metadata_attachment_name']:
                        metadata_dictionary = yaml_load(
                            stream=file_object.read()
                        )
                        logger.debug(
                            'Got metadata dictionary: %s',
                            metadata_dictionary
                        )
                        for metadata_name, value in metadata_dictionary.items():
                            metadata = MetadataType.objects.get(name=metadata_name)
                            self.document_metadata[metadata.pk] = value
                    else:
                        shared_uploaded_files.append(
                            SharedUploadedFile.objects.create(
                                file=file_object, filename=label
                            )
                        )
            else:
                # If it is not an attachment then it should be a body message
                # part. Another option is to use message.is_body().
                if message.detected_content_type == 'text/html':
                    label = 'email_body.html'
                else:
                    label = 'email_body.txt'

                if self.kwargs['store_body']:
                    with ContentFile(content=force_bytes(message.body), name=label) as file_object:
                        shared_uploaded_files.append(
                            SharedUploadedFile.objects.create(
                                file=file_object, filename=label
                            )
                        )

        return shared_uploaded_files

    def callback(self, document_file, **kwargs):
        for metadata_type_id, value in kwargs['document_metadata'].items():
            metadata_type = MetadataType.objects.get(pk=metadata_type_id)

            document_file.document.metadata.create(
                metadata_type=metadata_type, value=value
            )

    def clean(self):
        document_type = self.get_document_type()

        form_metadata_type = self.get_from_metadata_type()
        subject_metadata_type = self.get_subject_metadata_type()
        message_id_metadata_type = self.get_message_id_metadata_type()

        if form_metadata_type:
            if not document_type.metadata.filter(metadata_type=form_metadata_type).exists():
                raise ValidationError(
                    {
                        'from_metadata_type': _(
                            '"From" metadata type "%(metadata_type)s" is not '
                            'valid for the document type: %(document_type)s'
                        ) % {
                            'metadata_type': form_metadata_type,
                            'document_type': document_type
                        }
                    }
                )

        if subject_metadata_type:
            if not document_type.metadata.filter(metadata_type=subject_metadata_type).exists():
                raise ValidationError(
                    {
                        'subject_metadata_type': _(
                            'Subject metadata type "%(metadata_type)s" is not '
                            'valid for the document type: %(document_type)s'
                        ) % {
                            'metadata_type': subject_metadata_type,
                            'document_type': document_type
                        }
                    }
                )

        if message_id_metadata_type:
            if not document_type.metadata.filter(metadata_type=message_id_metadata_type).exists():
                raise ValidationError(
                    {
                        'message_id_metadata_type': _(
                            'Message ID metadata type "%(metadata_type)s" is not '
                            'valid for the document type: %(document_type)s'
                        ) % {
                            'metadata_type': subject_metadata_type,
                            'document_type': document_type
                        }
                    }
                )

    def get_callback_kwargs(self):
        callback_kwargs = super().get_callback_kwargs()
        callback_kwargs.update(
            {'document_metadata': self.document_metadata}
        )

        return callback_kwargs

    def get_from_metadata_type(self):
        pk = self.kwargs.get('from_metadata_type_id')

        if pk:
            return MetadataType.objects.get(pk=pk)

    def get_subject_metadata_type(self):
        pk = self.kwargs.get('subject_metadata_type_id')

        if pk:
            return MetadataType.objects.get(pk=pk)

    def get_message_id_metadata_type(self):
        pk = self.kwargs.get('message_id_metadata_type_id')

        if pk:
            return MetadataType.objects.get(pk=pk)


class SourceBackendIMAPEmail(
    SourceBackendCompressedMixin, SourceBackendEmailMixin,
    SourceBackendPeriodicMixin, SourceBaseMixin, SourceBackend
):
    field_order = (
        'mailbox', 'search_criteria', 'store_commands', 'execute_expunge',
        'mailbox_destination'
    )
    fields = {
        'mailbox': {
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_MAILBOX,
            'help_text': _('IMAP Mailbox from which to check for messages.'),
            'kwargs': {
                'max_length': 64,
            },
            'label': _('Mailbox')
        },
        'search_criteria': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_SEARCH_CRITERIA,
            'help_text': _(
                'Criteria to use when searching for messages to process. '
                'Use the format specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.4'
            ),
            'label': _('Search criteria'),
            'null': True,
        },
        'store_commands': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'default': DEFAULT_EMAIL_IMAP_STORE_COMMANDS,
            'help_text': _(
                'IMAP STORE command to execute on messages after they are '
                'processed. One command per line. Use the commands specified in '
                'https://tools.ietf.org/html/rfc2060.html#section-6.4.6 or '
                'the custom commands for your IMAP server.'
            ),
            'label': _('Store commands'),
            'null': True, 'required': False
        },
        'execute_expunge': {
            'class': 'django.forms.fields.BooleanField',
            'default': True,
            'help_text': _(
                'Execute the IMAP expunge command after processing each email '
                'message.'
            ),
            'label': _('Execute expunge'),
            'required': False
        },
        'mailbox_destination': {
            'blank': True,
            'class': 'django.forms.fields.CharField',
            'help_text': _(
                'IMAP Mailbox to which processed messages will be copied.'
            ),
            'label': _('Destination mailbox'),
            'max_length': 96,
            'null': True,
            'required': False
        }
    }
    label = _('IMAP email')
    uncompress_choices = SOURCE_INTERVAL_UNCOMPRESS_CHOICES
    widgets = {
        'search_criteria': {
            'class': 'django.forms.widgets.Textarea',
        },
        'store_commands': {
            'class': 'django.forms.widgets.Textarea',
        }
    }

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)

        logger.debug(msg='Starting IMAP email fetch')
        logger.debug('host: %s', self.kwargs['host'])
        logger.debug('ssl: %s', self.kwargs['ssl'])

        if self.kwargs['ssl']:
            imap_module_name = 'IMAP4_SSL'
        else:
            imap_module_name = 'IMAP4'

        imap_module = getattr(imaplib, imap_module_name)

        kwargs = {
            'host': self.kwargs['host'], 'port': self.kwargs['port']
        }

        with imap_module(**kwargs) as server:
            server.login(
                user=self.kwargs['username'], password=self.kwargs['password']
            )

            try:
                server.select(mailbox=self.kwargs['mailbox'])
            except Exception as exception:
                raise SourceException(
                    'Error selecting mailbox: {}; {}'.format(
                        self.kwargs['mailbox'], exception
                    )
                )
            else:
                try:
                    status, data = server.uid(
                        'SEARCH', None, *self.kwargs['search_criteria'].strip().split()
                    )
                except Exception as exception:
                    raise SourceException(
                        'Error executing search command; {}'.format(exception)
                    )
                else:
                    if data:
                        # data is a space separated sequence of message uids.
                        uids = data[0].split()
                        logger.debug('messages count: %s', len(uids))
                        logger.debug('message uids: %s', uids)

                        for uid in uids:
                            logger.debug('message uid: %s', uid)

                            try:
                                status, data = server.uid('FETCH', uid, '(RFC822)')
                            except Exception as exception:
                                raise SourceException(
                                    'Error fetching message uid: {}; {}'.format(
                                        uid, exception
                                    )
                                )
                            else:
                                try:
                                    shared_uploaded_files = self.process_message(
                                        message=data[0][1]
                                    )
                                except Exception as exception:
                                    raise SourceException(
                                        'Error processing message uid: {}; {}'.format(
                                            uid, exception
                                        )
                                    )
                                else:
                                    if not dry_run:
                                        if self.kwargs['store_commands']:
                                            for command in self.kwargs['store_commands'].split('\n'):
                                                try:
                                                    args = [uid]
                                                    args.extend(command.strip().split(' '))
                                                    server.uid('STORE', *args)
                                                except Exception as exception:
                                                    raise SourceException(
                                                        'Error executing IMAP store command "{}" '
                                                        'on message uid {}; {}'.format(
                                                            command, uid, exception
                                                        )
                                                    )

                                        if self.kwargs['mailbox_destination']:
                                            try:
                                                server.uid(
                                                    'COPY', uid, self.kwargs['mailbox_destination']
                                                )
                                            except Exception as exception:
                                                raise SourceException(
                                                    'Error copying message uid {} to mailbox {}; '
                                                    '{}'.format(
                                                        uid, self.kwargs['mailbox_destination'], exception
                                                    )
                                                )

                                        if self.kwargs['execute_expunge']:
                                            server.expunge()

                                    return shared_uploaded_files


class SourceBackendPOP3Email(
    SourceBackendCompressedMixin, SourceBackendEmailMixin,
    SourceBackendPeriodicMixin, SourceBaseMixin, SourceBackend
):
    field_order = ('timeout', 'uncompress',)
    fields = {
        'timeout': {
            'class': 'django.forms.fields.IntegerField',
            'default': DEFAULT_EMAIL_POP3_TIMEOUT,
            'kwargs': {
                'min_value': 0
            },
            'label': _('Timeout')
        }
    }
    label = _('POP3 email')

    def get_shared_uploaded_files(self):
        dry_run = self.process_kwargs.get('dry_run', False)

        logger.debug(msg='Starting POP3 email fetch')
        logger.debug('host: %s', self.kwargs['host'])
        logger.debug('ssl: %s', self.kwargs['ssl'])

        if self.kwargs['ssl']:
            pop3_module_name = 'POP3_SSL'
        else:
            pop3_module_name = 'POP3'

        pop3_module = getattr(poplib, pop3_module_name)

        kwargs = {
            'host': self.kwargs['host'], 'port': self.kwargs['port'],
            'timeout': self.kwargs['timeout']
        }

        server = pop3_module(**kwargs)
        try:
            server.getwelcome()
            server.user(self.kwargs['username'])
            server.pass_(self.kwargs['password'])

            messages_info = server.list()

            logger.debug(msg='messages_info:')
            logger.debug(msg=messages_info)
            logger.debug('messages count: %s', len(messages_info[1]))

            for message_info in messages_info[1]:
                message_number, message_size = message_info.split()
                message_number = int(message_number)

                logger.debug('message_number: %s', message_number)
                logger.debug('message_size: %s', message_size)

                message_lines = server.retr(which=message_number)[1]
                message_complete = force_text(s=b'\n'.join(message_lines))

                shared_uploaded_files = self.process_message(
                    message=message_complete
                )
                if not dry_run:
                    server.dele(which=message_number)

                return shared_uploaded_files
        finally:
            server.quit()



from django.core.exceptions import ValidationError

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.metadata.models import MetadataType

from ..events import event_source_created
from ..models import Source
from ..permissions import permission_sources_create, permission_sources_edit

from .literals import TEST_EMAIL_BASE64_FILENAME
from .mixins.email_source_mixins import (
    EmailSourceBackendTestMixin, EmailSourceBackendViewTestMixin
)


class EmailSourceViewTestCase(
    EmailSourceBackendTestMixin, EmailSourceBackendViewTestMixin,
    GenericDocumentViewTestCase
):
    _test_email_source_content = TEST_EMAIL_BASE64_FILENAME
    auto_upload_test_document = False

    def test_email_source_create_view(self):
        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_email_source_create_view()

        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_metadata_type_validation_invalid_from(self):
        self._silence_logger(name='mayan.apps.logging.middleware')
        self._silence_logger(name='mayan.apps.navigation.classes')

        test_metadata_type = MetadataType.objects.create(
            name='test_metadata_type'
        )

        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=ValidationError):
            response = self._request_test_email_source_create_view(
                extra_data={
                    'from_metadata_type_id': test_metadata_type.pk,
                }
            )
            self.assertEqual(response.status_code, 200)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_validation_valid_from(self):
        test_metadata_type = MetadataType.objects.create(
            name='test_metadata_type'
        )

        self.test_document_type.metadata.create(
            metadata_type=test_metadata_type
        )

        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_email_source_create_view(
            extra_data={
                'from_metadata_type_id': test_metadata_type.pk,
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_metadata_type_validation_invalid_subject(self):
        self._silence_logger(name='mayan.apps.logging.middleware')
        self._silence_logger(name='mayan.apps.navigation.classes')

        test_metadata_type = MetadataType.objects.create(
            name='test_metadata_type'
        )

        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        with self.assertRaises(expected_exception=ValidationError):
            response = self._request_test_email_source_create_view(
                extra_data={
                    'subject_metadata_type_id': test_metadata_type.pk
                }
            )
            self.assertEqual(response.status_code, 200)

        self.assertEqual(Source.objects.count(), source_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_metadata_type_validation_valid_subject(self):
        test_metadata_type = MetadataType.objects.create(
            name='test_metadata_type'
        )

        self.test_document_type.metadata.create(
            metadata_type=test_metadata_type
        )

        self.grant_permission(permission=permission_sources_create)

        source_count = Source.objects.count()

        self._clear_events()

        response = self._request_test_email_source_create_view(
            extra_data={
                'subject_metadata_type_id': test_metadata_type.pk
            }
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Source.objects.count(), source_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_email_source_test_view_no_permission(self):
        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_email_source_test_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter')

        self.grant_access(
            obj=self.test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_file)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0003_notification'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='EventType',
            new_name='StoredEventType',
        ),
        migrations.AlterModelOptions(
            name='storedeventtype',
            options={
                'verbose_name': 'Stored event type',
                'verbose_name_plural': 'Stored event types'
            },
        ),
        migrations.RemoveField(
            model_name='eventsubscription',
            name='event_type',
        ),
        migrations.AddField(
            model_name='eventsubscription',
            name='stored_event_type',
            field=models.ForeignKey(
                default=1, on_delete=django.db.models.deletion.CASCADE,
                related_name='event_subscriptions',
                to='events.StoredEventType', verbose_name='Event type'
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='eventsubscription',
            name='user',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='event_subscriptions',
                to=settings.AUTH_USER_MODEL, verbose_name='User'
            ),
        ),
    ]



import logging

from django.apps import apps

from mayan.celery import app

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_launch_all_workflows():
    Document = apps.get_model(app_label='documents', model_name='Document')
    Workflow = apps.get_model(
        app_label='document_states', model_name='Workflow'
    )

    logger.info('Start launching workflows')
    for document in Document.valid.all():
        Workflow.objects.launch_for(document=document)

    logger.info('Finished launching workflows')


@app.task(ignore_result=True)
def task_launch_workflow(workflow_id):
    Document = apps.get_model(app_label='documents', model_name='Document')
    Workflow = apps.get_model(
        app_label='document_states', model_name='Workflow'
    )

    workflow = Workflow.objects.get(pk=workflow_id)

    logger.info('Start launching workflow: %d', workflow_id)
    for document in Document.valid.filter(document_type__in=workflow.document_types.all()):
        workflow.launch_for(document=document)

    logger.info('Finished launching workflow: %d', workflow_id)


@app.task(ignore_result=True)
def task_launch_workflow_for(document_id, workflow_id):
    Document = apps.get_model(app_label='documents', model_name='Document')
    Workflow = apps.get_model(
        app_label='document_states', model_name='Workflow'
    )

    document = Document.valid.get(pk=document_id)
    workflow = Workflow.objects.get(pk=workflow_id)

    logger.info(
        'Start launching workflow: %d for document: %d',
        workflow_id, document_id
    )
    workflow.launch_for(document=document)

    logger.info(
        'Finished launching workflow: %d for document: %d', workflow_id,
        document_id
    )


@app.task(ignore_result=True)
def task_launch_all_workflow_for(document_id):
    Document = apps.get_model(app_label='documents', model_name='Document')
    Workflow = apps.get_model(
        app_label='document_states', model_name='Workflow'
    )

    document = Document.valid.get(pk=document_id)

    logger.info(
        'Start launching all workflows for document: %d', document_id
    )
    Workflow.objects.launch_for(document=document)

    logger.info(
        'Finished launching all workflows for document: %d', document_id
    )



class ACLsBaseException(Exception):
    """
    Base exception for the acls app
    """


class PermissionNotValidForClass(ACLsBaseException):
    """
    The permission is not one that has been registered for a class using the
    ModelPermission class.
    """



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..models import IndexInstanceNode
from ..permissions import permission_index_template_rebuild

from .mixins import (
    IndexInstanceTestMixin, IndexTemplateTestMixin, IndexToolsViewTestMixin,
    IndexTemplateViewTestMixin
)


class IndexToolsViewTestCase(
    IndexInstanceTestMixin, IndexTemplateTestMixin,
    IndexTemplateViewTestMixin, IndexToolsViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        IndexInstanceNode.objects.exclude(parent=None).delete()

    def test_index_all_rebuild_get_view_no_permission(self):
        index_instance_node_count = self.test_index_instance.get_descendants().count()

        response = self._request_index_all_rebuild_get_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_index_template.label
        )

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_rebuild_post_view_no_permission(self):
        index_instance_node_count = self.test_index_instance.get_descendants().count()

        response = self._request_indexes_rebuild_post_view()
        # No error since we just don't see the index.
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_rebuild_get_view_with_access(self):
        index_instance_node_count = self.test_index_instance.get_descendants().count()

        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        response = self._request_index_all_rebuild_get_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_index_template.label
        )

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_rebuild_post_view_with_access(self):
        index_instance_node_count = self.test_index_instance.get_descendants().count()

        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        response = self._request_indexes_rebuild_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_reset_get_view_no_permission(self):
        self.test_index_template.rebuild()

        index_instance_node_count = self.test_index_instance.get_descendants().count()

        response = self._request_index_all_reset_get_view()
        self.assertNotContains(
            response=response, status_code=200,
            text=self.test_index_template.label
        )

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_reset_post_view_no_permission(self):
        self.test_index_template.rebuild()

        index_instance_node_count = self.test_index_instance.get_descendants().count()

        response = self._request_index_all_reset_post_view()
        # No error since we just don't see the index.
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_reset_get_view_with_access(self):
        self.test_index_template.rebuild()

        index_instance_node_count = self.test_index_instance.get_descendants().count()

        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        response = self._request_index_all_reset_get_view()
        self.assertContains(
            response=response, status_code=200,
            text=self.test_index_template.label
        )

        self.assertEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )

    def test_index_all_reset_post_view_with_access(self):
        self.test_index_template.rebuild()

        index_instance_node_count = self.test_index_instance.get_descendants().count()

        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        response = self._request_index_all_reset_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            self.test_index_instance.get_descendants().count(),
            index_instance_node_count
        )



import json
import logging

from django import forms
from django.apps import apps
from django.utils.encoding import force_text
from django.utils.translation import ugettext, ugettext_lazy as _

from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.tasks import task_document_file_upload

from ..classes import DocumentCreateWizardStep
from ..tasks import task_process_document_upload

from .literals import (
    DEFAULT_PERIOD_INTERVAL, SOURCE_INTERACTIVE_UNCOMPRESS_CHOICES,
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK
)

logger = logging.getLogger(name=__name__)


class SourceBaseMixin:
    def callback(self, document_file, **kwargs):
        return

    def clean(self):
        return

    def get_callback_kwargs(self):
        return {}

    def get_document(self):
        raise NotImplementedError

    def get_document_description(self):
        return None

    def get_document_file_action(self):
        return None

    def get_document_file_comment(self):
        return None

    def get_document_label(self):
        return None

    def get_document_language(self):
        return None

    def get_document_type(self):
        raise NotImplementedError

    def get_task_extra_kwargs(self):
        return {}

    def get_user(self):
        return None

    def process_document_file(self, **kwargs):
        self.process_kwargs = kwargs

        document = self.get_document()
        user = self.get_user()

        if user:
            user_id = user.pk
        else:
            user_id = None

        for self.shared_uploaded_file in self.get_shared_uploaded_files() or ():
            # Call the hooks here too as in the model for early detection and
            # exception raise when using the views.
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': document,
                    'file_object': self.shared_uploaded_file,
                    'user': user
                }
            )

            kwargs = {
                'action': self.get_document_file_action(),
                'comment': self.get_document_file_comment(),
                'document_id': document.pk,
                'shared_uploaded_file_id': self.shared_uploaded_file.pk,
                'user_id': user_id
            }

            kwargs.update(self.get_task_extra_kwargs())

            task_document_file_upload.apply_async(kwargs=kwargs)

    def process_documents(self, **kwargs):
        self.process_kwargs = kwargs

        document_type = self.get_document_type()
        user = self.get_user()

        if user:
            user_id = user.pk
        else:
            user_id = None

        for self.shared_uploaded_file in self.get_shared_uploaded_files() or ():
            kwargs = {
                'callback_kwargs': self.get_callback_kwargs(),
                'description': self.get_document_description(),
                'document_type_id': document_type.pk,
                'label': self.get_document_label(),
                'language': self.get_document_language(),
                'shared_uploaded_file_id': self.shared_uploaded_file.pk,
                'source_id': self.model_instance_id,
                'user_id': user_id
            }
            kwargs.update(self.get_task_extra_kwargs())

            task_process_document_upload.apply_async(kwargs=kwargs)


class SourceBackendCompressedMixin:
    uncompress_choices = SOURCE_INTERACTIVE_UNCOMPRESS_CHOICES

    @classmethod
    def get_setup_form_schema(cls):
        result = super().get_setup_form_schema()

        result['fields'].update(
            {
                'uncompress': {
                    'label': _('Uncompress'),
                    'class': 'django.forms.ChoiceField',
                    'default': SOURCE_UNCOMPRESS_CHOICE_ASK,
                    'help_text': _(
                        'Whether to expand or not compressed archives.'
                    ), 'kwargs': {
                        'choices': cls.uncompress_choices,
                    }, 'required': True
                }
            }
        )
        result['field_order'] = ('uncompress',) + result['field_order']

        result['widgets'].update(
            {
                'uncompress': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'},
                    }
                }
            }
        )
        return result

    @classmethod
    def get_upload_form_class(cls):
        class CompressedSourceUploadForm(super().get_upload_form_class()):
            expand = forms.BooleanField(
                label=_('Expand compressed files'), required=False,
                help_text=ugettext(
                    'Upload a compressed file\'s contained files as '
                    'individual documents.'
                )
            )

            def __init__(self, *args, **kwargs):
                self.field_order = ['expand']
                super().__init__(*args, **kwargs)

        return CompressedSourceUploadForm

    def get_expand(self):
        if self.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ASK:
            return self.process_kwargs['forms']['source_form'].cleaned_data.get('expand')
        else:
            if self.kwargs['uncompress'] == SOURCE_UNCOMPRESS_CHOICE_ALWAYS:
                return True
            else:
                return False

    def get_task_extra_kwargs(self):
        return {'expand': self.get_expand()}


class SourceBackendInteractiveMixin:
    is_interactive = True

    def callback(self, document_file, **kwargs):
        DocumentCreateWizardStep.post_upload_process(
            document=document_file.document,
            query_string=kwargs.get('query_string', '')
        )

    def get_callback_kwargs(self):
        query_string = ''

        query_dict = self.process_kwargs['request'].GET.copy()
        query_dict.update(self.process_kwargs['request'].POST)

        # Convert into a string. Make sure it is a QueryDict object from a
        # request and not just a simple dictionary.
        if hasattr(query_dict, 'urlencode'):
            query_string = query_dict.urlencode()

        return {
            'query_string': query_string
        }

    def get_document(self):
        return self.process_kwargs['document']

    def get_document_description(self):
        return self.process_kwargs['forms']['document_form'].cleaned_data.get('description')

    def get_document_file_action(self):
        return int(self.process_kwargs['forms']['document_form'].cleaned_data.get('action'))

    def get_document_file_comment(self):
        return self.process_kwargs['forms']['document_form'].cleaned_data.get('comment')

    def get_document_label(self):
        return self.process_kwargs['forms']['document_form'].get_final_label(
            filename=force_text(self.shared_uploaded_file)
        )

    def get_document_language(self):
        return self.process_kwargs['forms']['document_form'].cleaned_data.get('language')

    def get_document_type(self):
        return self.process_kwargs['document_type']

    def get_user(self):
        if not self.process_kwargs['request'].user.is_anonymous:
            return self.process_kwargs['request'].user
        else:
            return None


class SourceBackendPeriodicMixin:
    @classmethod
    def get_setup_form_schema(cls):
        result = super().get_setup_form_schema()

        result['fields'].update(
            {
                'document_type_id': {
                    'class': 'django.forms.ChoiceField',
                    'default': '',
                    'help_text': _(
                        'Assign a document type to documents uploaded from this '
                        'source.'
                    ),
                    'kwargs': {
                        'choices': [(document_type.id, document_type) for document_type in DocumentType.objects.all()],
                    },
                    'label': _('Document type'),
                    'required': True
                },
                'interval': {
                    'class': 'django.forms.IntegerField',
                    'default': DEFAULT_PERIOD_INTERVAL,
                    'help_text': _(
                        'Interval in seconds between checks for new '
                        'documents.'
                    ),
                    'kwargs': {
                        'min_value': 0
                    },
                    'label': _('Interval'),
                    'required': True
                },
            }
        )
        result['field_order'] = ('document_type_id', 'interval',) + result['field_order']

        result['widgets'].update(
            {
                'document_type_id': {
                    'class': 'django.forms.widgets.Select', 'kwargs': {
                        'attrs': {'class': 'select2'},
                    }
                }
            }
        )

        return result

    def create(self):
        IntervalSchedule = apps.get_model(
            app_label='django_celery_beat', model_name='IntervalSchedule'
        )
        PeriodicTask = apps.get_model(
            app_label='django_celery_beat', model_name='PeriodicTask'
        )

        # Create a new interval or use an existing one
        interval_instance, created = IntervalSchedule.objects.get_or_create(
            every=self.kwargs['interval'], period='seconds'
        )

        PeriodicTask.objects.create(
            name=self.get_periodic_task_name(),
            interval=interval_instance,
            task='mayan.apps.sources.tasks.task_source_process_document',
            kwargs=json.dumps(obj={'source_id': self.model_instance_id})
        )

    def delete(self):
        self.delete_periodic_task(pk=self.model_instance_id)

    def get_document_type(self):
        return DocumentType.objects.get(pk=self.kwargs['document_type_id'])

    def delete_periodic_task(self, pk=None):
        PeriodicTask = apps.get_model(
            app_label='django_celery_beat', model_name='PeriodicTask'
        )

        try:
            periodic_task = PeriodicTask.objects.get(
                name=self.get_periodic_task_name(pk=pk)
            )

            interval_instance = periodic_task.interval

            if tuple(interval_instance.periodictask_set.values_list('id', flat=True)) == (periodic_task.pk,):
                # Only delete the interval if nobody else is using it.
                interval_instance.delete()
            else:
                periodic_task.delete()
        except PeriodicTask.DoesNotExist:
            logger.warning(
                'Tried to delete non existent periodic task "%s"',
                self.get_periodic_task_name(pk=pk)
            )

    def get_periodic_task_name(self, pk=None):
        return 'check_interval_source-{}'.format(pk or self.model_instance_id)

    def save(self):
        self.delete_periodic_task()
        self.create()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0071_auto_20201128_0330'),
    ]

    operations = [
        migrations.CreateModel(
            name='RecentlyCreatedDocument',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('documents.document',),
        ),
        migrations.AlterModelOptions(
            name='favoritedocument',
            options={
                'ordering': ('datetime_added',),
                'verbose_name': 'Favorite document',
                'verbose_name_plural': 'Favorite documents'
            },
        ),
        migrations.AddField(
            model_name='favoritedocument',
            name='datetime_added',
            field=models.DateTimeField(
                auto_now=True, db_index=True,
                verbose_name='Date and time added'
            ),
        ),
        migrations.AlterField(
            model_name='document',
            name='datetime_created',
            field=models.DateTimeField(
                auto_now_add=True, db_index=True,
                help_text='The server date and time when the document '
                'was finally processed and created in the system.',
                verbose_name='Created'
            ),
        ),
    ]



from rest_framework import status
from rest_framework.response import Response

from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.rest_api import generics

from .models import Source
from .permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)
from .serializers import SourceBackendActionSerializer, SourceSerializer


class APISourceListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the source.
    post: Create a new source.
    """
    mayan_object_permissions = {
        'GET': (permission_sources_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_sources_create,)
    }
    queryset = Source.objects.all()
    serializer_class = SourceSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISourceView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected source.
    get: Details of the selected source.
    patch: Edit the selected source.
    put: Edit the selected source.
    """
    lookup_url_kwarg = 'source_id'
    mayan_object_permissions = {
        'DELETE': (permission_sources_delete,),
        'GET': (permission_sources_view,),
        'PATCH': (permission_sources_edit,),
        'PUT': (permission_sources_edit,)
    }
    queryset = Source.objects.all()
    serializer_class = SourceSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APISourceActionView(generics.ObjectActionAPIView):
    """
    get: Get data from a source action.
    post: Execute a source action.
    """
    lookup_url_kwarg = 'source_id'
    mayan_object_permissions = {
        'GET': (permission_document_create,),
        'POST': (permission_document_create,)
    }
    serializer_class = SourceBackendActionSerializer

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request=request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        if self.get_action().confirmation:
            handler = self.http_method_not_allowed
            response = handler(request, *args, **kwargs)
            self.response = self.finalize_response(
                request, response, *args, **kwargs
            )
            return self.response
        else:
            return self.view_action(request, *args, **kwargs)

    def get_action(self):
        return self.object.get_action(name=self.kwargs['action_name'])

    def get_queryset(self):
        return Source.objects.filter(enabled=True)

    def object_action(self, request, serializer):
        query_dict = request.GET

        arguments = serializer.data.get('arguments', {}) or {}
        arguments.update(query_dict)

        return self.object.execute_action(
            name=self.kwargs['action_name'], request=request, **arguments
        ) or (None, None)

    def post(self, request, *args, **kwargs):
        if not self.get_action().confirmation:
            handler = self.http_method_not_allowed
            response = handler(request, *args, **kwargs)
            self.response = self.finalize_response(
                request, response, *args, **kwargs
            )
            return self.response
        else:
            return self.view_action(request, *args, **kwargs)

    def view_action(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(self.object, key, value)

        data, response = self.object_action(
            request=request, serializer=serializer
        )

        if response:
            return response
        else:
            if data:
                # If object action returned serializer.data.
                headers = self.get_success_headers(data=data)
                return Response(
                    data=data, status=status.HTTP_200_OK, headers=headers
                )
            else:
                return Response(status=status.HTTP_200_OK)



from django.core import management
from django.utils.translation import ugettext_lazy as _

from ...classes import GoogleFontDependency, JavaScriptDependency


class Command(management.BaseCommand):
    help = 'Install JavaScript dependencies.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--app', action='store', dest='app',
            help=_('Process a specific app.'),
        )
        parser.add_argument(
            '--force', action='store_true', dest='force',
            help=_('Force installation even if already installed.'),
        )

    def handle(self, *args, **options):
        JavaScriptDependency.install_multiple(
            app_label=options['app'], force=options['force'],
            subclass_only=True
        )
        GoogleFontDependency.install_multiple(
            app_label=options['app'], force=options['force'],
            subclass_only=True
        )



from django.shortcuts import get_object_or_404

from mayan.apps.views.http import URL

from .models import DocumentMetadata, MetadataType


def decode_metadata_from_query_string(query_string=None):
    """
    Parse a URL query string to a list of metadata
    """
    metadata_dict = {
        'metadata_type_id': {},
        'value': {}
    }
    metadata_list = []
    if query_string:
        # Match out of order metadata_type ids with metadata values from
        # request.
        for key, value in URL(query_string=query_string).args.items():
            if key.startswith('metadata'):
                index, element = key.split('_', 1)
                index = index[8:]
                metadata_dict[element][index] = value

        # Convert the nested dictionary into a list of id+values dictionaries
        for order, identifier in metadata_dict['metadata_type_id'].items():
            if order in metadata_dict['value'].keys():
                metadata_list.append(
                    {
                        'metadata_type_id': identifier,
                        'value': metadata_dict['value'][order]
                    }
                )

    return metadata_list


def save_metadata_list(metadata_list, document, create=False, _user=None):
    """
    Take a list of metadata dictionaries and associate them to a
    document
    """
    for item in metadata_list:
        save_metadata(
            metadata_dict=item, document=document, create=create, _user=_user
        )


def save_metadata(metadata_dict, document, create=False, _user=None):
    """
    Take a dictionary of metadata type & value and associate it to a
    document
    """
    parameters = {
        'document': document,
        'metadata_type': get_object_or_404(
            klass=MetadataType, pk=metadata_dict['metadata_type_id']
        )
    }

    if create:
        # Use matched metadata now to create document metadata.
        try:
            DocumentMetadata.objects.get(**parameters)
        except DocumentMetadata.DoesNotExist:
            document_metadata = DocumentMetadata(**parameters)
            document_metadata._event_actior = _user
            document_metadata.save()
    else:
        try:
            document_metadata = DocumentMetadata.objects.get(
                document=document,
                metadata_type=get_object_or_404(
                    klass=MetadataType, pk=metadata_dict['metadata_type_id']
                ),
            )
        except DocumentMetadata.DoesNotExist:
            document_metadata = None

    if document_metadata:
        document_metadata.value = metadata_dict['value']
        document_metadata._event_actor = _user
        document_metadata.save()


def set_bulk_metadata(document, metadata_dictionary):
    document_type_metadata_types = document.document_type.metadata.values_list(
        'metadata_type', flat=True
    )

    for metadata_type_name, value in metadata_dictionary.items():
        metadata_type = MetadataType.objects.get(name=metadata_type_name)

        if document_type_metadata_types.filter(metadata_type=metadata_type).exists():
            DocumentMetadata.objects.get_or_create(
                document=document, metadata_type=metadata_type, value=value
            )



import logging

from django.contrib import messages
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.views.mixins import ExternalObjectViewMixin

from ..forms import NewDocumentForm
from ..models import Source

from .base import UploadBaseView

__all__ = ('DocumentUploadInteractiveView',)
logger = logging.getLogger(name=__name__)


class DocumentUploadInteractiveView(ExternalObjectViewMixin, UploadBaseView):
    external_object_class = DocumentType
    external_object_permission = permission_document_create
    document_form = NewDocumentForm
    object_permission = permission_document_create

    def forms_valid(self, forms):
        source_backend_instance = self.source.get_backend_instance()

        try:
            source_backend_instance.process_documents(
                document_type=self.external_object, forms=forms,
                request=self.request
            )
        except Exception as exception:
            message = _(
                'Error processing source document upload; '
                '%(exception)s'
            ) % {
                'exception': exception,
            }
            logger.critical(msg=message, exc_info=True)
            if self.request.is_ajax():
                return JsonResponse(
                    data={'error': force_text(s=message)}, status=500
                )
            else:
                messages.error(
                    message=message.replace('\n', ' '),
                    request=self.request
                )
                raise type(exception)(message)
        else:
            messages.success(
                message=_(
                    'New document queued for upload and will be '
                    'available shortly.'
                ), request=self.request
            )

        return HttpResponseRedirect(
            redirect_to='{}?{}'.format(
                reverse(
                    viewname=self.request.resolver_match.view_name,
                    kwargs=self.request.resolver_match.kwargs
                ), self.request.META['QUERY_STRING']
            ),
        )

    def get_active_tab_links(self):
        sources = AccessControlList.objects.restrict_queryset(
            permission=permission_document_create,
            queryset=Source.objects.interactive().filter(enabled=True),
            user=self.request.user
        )
        return [
            UploadBaseView.get_tab_link_for_source(source=source)
            for source in sources
        ]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'document_type': self.external_object,
                'title': _(
                    'Upload a document of type "%(document_type)s" from '
                    'source: %(source)s'
                ) % {
                    'document_type': self.external_object,
                    'source': self.source.label
                }
            }
        )

        return context

    def get_form_extra_kwargs__document_form(self):
        return {
            'document_type': self.external_object
        }

    def get_form_extra_kwargs__source_form(self):
        return {
            'source': self.source
        }

    def get_pk_url_kwargs(self):
        return {
            'pk': self.request.GET.get(
                'document_type_id', self.request.POST.get('document_type_id')
            )
        }



class DynamicSearchException(Exception):
    """
    Base exception for the app.
    """


class DynamicSearchRetry(DynamicSearchException):
    """
    Exception to encapsulate backend specific errors that should be retried.
    """



from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import QuotaTestMixin


class QuotaModelTestCase(QuotaTestMixin, BaseTestCase):
    def test_method_get_absolute_url(self):
        self._create_test_quota()

        self.assertTrue(self.test_quota.get_absolute_url())



from django.contrib import admin

from .models import Message


@admin.register(Message)
class MessageAdmin(admin.ModelAdmin):
    list_display = ('date_time', 'sender_object', 'user', 'subject', 'read')



from ..classes import SourceBackend
from ..source_backends.email_backends import SourceBackendEmailMixin
from ..source_backends.mixins import (
    SourceBaseMixin, SourceBackendPeriodicMixin
)

__all__ = (
    'SourceBackendSimple', 'SourceBackendTestPeriodic',
    'SourceBackendTestEmail'
)


class SourceBackendSimple(SourceBaseMixin, SourceBackend):
    label = 'Test source backend'

    def process_documents(self, dry_run=False):
        """Do nothing. This method is added to allow view testing."""


class SourceBackendTestPeriodic(
    SourceBackendPeriodicMixin, SourceBaseMixin, SourceBackend
):
    label = 'Test periodic source backend'


class SourceBackendTestEmail(
    SourceBackendEmailMixin, SourceBackendPeriodicMixin, SourceBaseMixin,
    SourceBackend
):
    label = 'Test email source backend'

    def get_shared_uploaded_files(self):
        data = self.get_model_instance().get_backend_data()

        message = getattr(self, 'content', data.get('_test_content'))

        return self.process_message(message=message)



import json

from django.db import models


class SourceManager(models.Manager):
    def create_backend(self, label, backend_path, backend_data=None):
        self.create(
            backend_path=backend_path,
            backend_data=json.dumps(obj=backend_data or {}),
            label=label
        )

    def interactive(self):
        interactive_sources_ids = []
        for source in self.all():
            if getattr(source.get_backend(), 'is_interactive', False):
                interactive_sources_ids.append(source.pk)

        return self.filter(id__in=interactive_sources_ids)



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.sources.tests.mixins.web_form_source_mixins import WebFormSourceTestMixin

from ..events import event_tag_attached

from .mixins import TagTestMixin, TaggedDocumentUploadWizardStepViewTestMixin


class TaggedDocumentUploadViewTestCase(
    TaggedDocumentUploadWizardStepViewTestMixin, TagTestMixin,
    WebFormSourceTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_upload_interactive_view_with_full_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_upload_interactive_document_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(self.test_tag in Document.objects.first().tags.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 6)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self.test_tag)
        self.assertEqual(events[5].actor, test_document)
        self.assertEqual(events[5].target, test_document)
        self.assertEqual(events[5].verb, event_tag_attached.id)

    def test_upload_interactive_multiple_tags_view_full_access(self):
        self._create_test_tag()
        self._create_test_tag()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_upload_interactive_document_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_tags[0] in Document.objects.first().tags.all()
        )
        self.assertTrue(
            self.test_tags[1] in Document.objects.first().tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 7)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self.test_tags[0])
        self.assertEqual(events[5].actor, test_document)
        self.assertEqual(events[5].target, test_document)
        self.assertEqual(events[5].verb, event_tag_attached.id)

        self.assertEqual(events[6].action_object, self.test_tags[1])
        self.assertEqual(events[6].actor, test_document)
        self.assertEqual(events[6].target, test_document)
        self.assertEqual(events[6].verb, event_tag_attached.id)



import io
import logging
import shutil
import struct

from PIL import Image
import PyPDF2
import sh

from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import NamedTemporaryFile

from ..classes import ConverterBase
from ..exceptions import PageCountError
from ..settings import setting_graphics_backend_arguments

from ..literals import (
    DEFAULT_PDFTOPPM_DPI, DEFAULT_PDFTOPPM_FORMAT, DEFAULT_PDFTOPPM_PATH,
    DEFAULT_PDFINFO_PATH, DEFAULT_PILLOW_MAXIMUM_IMAGE_PIXELS
)

logger = logging.getLogger(name=__name__)
pdftoppm_path = setting_graphics_backend_arguments.value.get(
    'pdftoppm_path', DEFAULT_PDFTOPPM_PATH
)

try:
    pdftoppm = sh.Command(path=pdftoppm_path)
except sh.CommandNotFound:
    pdftoppm = None
else:
    pdftoppm_format = '-{}'.format(
        setting_graphics_backend_arguments.value.get(
            'pdftoppm_format', DEFAULT_PDFTOPPM_FORMAT
        )
    )

    pdftoppm_dpi = format(
        setting_graphics_backend_arguments.value.get(
            'pdftoppm_dpi', DEFAULT_PDFTOPPM_DPI
        )
    )

    pdftoppm = pdftoppm.bake(pdftoppm_format, '-r', pdftoppm_dpi)

pdfinfo_path = setting_graphics_backend_arguments.value.get(
    'pdfinfo_path', DEFAULT_PDFINFO_PATH
)

try:
    pdfinfo = sh.Command(path=pdfinfo_path)
except sh.CommandNotFound:
    pdfinfo = None


pillow_maximum_image_pixels = setting_graphics_backend_arguments.value.get(
    'pillow_maximum_image_pixels', DEFAULT_PILLOW_MAXIMUM_IMAGE_PIXELS
)
Image.MAX_IMAGE_PIXELS = pillow_maximum_image_pixels


class Python(ConverterBase):
    def convert(self, *args, **kwargs):
        super().convert(*args, **kwargs)

        if self.mime_type == 'application/pdf' and pdftoppm:
            with NamedTemporaryFile() as new_file_object:
                self.file_object.seek(0)
                shutil.copyfileobj(fsrc=self.file_object, fdst=new_file_object)
                self.file_object.seek(0)
                new_file_object.seek(0)

                image_buffer = io.BytesIO()
                pdftoppm(
                    new_file_object.name, f=self.page_number + 1,
                    l=self.page_number + 1, _out=image_buffer
                )
                image_buffer.seek(0)
                return Image.open(fp=image_buffer)

    def get_page_count(self):
        super().get_page_count()

        page_count = 1

        if self.mime_type == 'application/pdf' or self.soffice_file:
            if self.soffice_file:
                file_object = self.soffice_file
            else:
                file_object = self.file_object

            try:
                # Try PyPDF to determine the page number
                pdf_reader = PyPDF2.PdfFileReader(
                    stream=file_object, strict=False
                )
                page_count = pdf_reader.getNumPages()
            except Exception as exception:
                if force_text(s=exception) == 'File has not been decrypted':
                    # File is encrypted, try to decrypt using a blank
                    # password.
                    file_object.seek(0)
                    pdf_reader = PyPDF2.PdfFileReader(
                        stream=file_object, strict=False
                    )
                    try:
                        pdf_reader.decrypt(password=b'')
                        page_count = pdf_reader.getNumPages()
                    except Exception as exception:
                        file_object.seek(0)
                        if force_text(s=exception) == 'only algorithm code 1 and 2 are supported':
                            # PDF uses an unsupported encryption
                            # Try poppler-util's pdfinfo
                            page_count = self.get_pdfinfo_page_count(file_object)
                            return page_count
                        else:
                            error_message = _(
                                'Exception determining PDF page count; %s'
                            ) % exception
                            logger.error(error_message, exc_info=True)
                            raise PageCountError(error_message)
                elif force_text(s=exception) == 'EOF marker not found':
                    # PyPDF2 issue: https://github.com/mstamy2/PyPDF2/issues/177
                    # Try poppler-util's pdfinfo
                    logger.debug('PyPDF2 GitHub issue #177 : EOF marker not found')
                    file_object.seek(0)
                    page_count = self.get_pdfinfo_page_count(file_object)
                    return page_count
                else:
                    error_message = _(
                        'Exception determining PDF page count; %s'
                    ) % exception
                    logger.error(error_message, exc_info=True)
                    raise PageCountError(error_message)
            else:
                logger.debug('Document contains %d pages', page_count)
                return page_count
            finally:
                file_object.seek(0)
        else:
            try:
                image = Image.open(fp=self.file_object)
            except IOError as exception:
                error_message = _(
                    'Exception determining page count using Pillow; %s'
                ) % exception
                logger.error(error_message)
                raise PageCountError(error_message)
            finally:
                self.file_object.seek(0)

            # Get total page count by attempting to seek to an increasing
            # page count number until an EOFError or struct.error exception
            # are raised.
            try:
                while True:
                    image.seek(image.tell() + 1)
                    page_count += 1
            except EOFError:
                """End of sequence"""
            except struct.error:
                """
                struct.error was raise for a TIFF file converted to JPEG
                GitLab issue #767 "Upload Error: unpack_from requires a
                buffer of at least 2 bytes"
                """
                logger.debug('image page count detection raised struct.error')

            return page_count

    def get_pdfinfo_page_count(self, file_object):
        process = pdfinfo('-', _in=file_object)
        page_count = int(
            list(filter(
                lambda line: line.startswith('Pages:'),
                force_text(s=process.stdout).split('\n')
            ))[0].replace('Pages:', '')
        )
        file_object.seek(0)
        logger.debug(
            'Document contains %d pages', page_count
        )
        return page_count



from mayan.apps.converter.exceptions import AppImageError
from mayan.apps.events.classes import EventManagerMethodAfter
from mayan.apps.events.decorators import method_event

from ..events import event_trashed_document_restored
from ..literals import IMAGE_ERROR_NO_VERSION_PAGES
from ..managers import TrashCanManager

from .document_models import Document

__all__ = ('TrashedDocument',)


class TrashedDocument(Document):
    objects = TrashCanManager()

    class Meta:
        proxy = True

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        first_page = self.pages.first()
        if first_page:
            return first_page.get_api_image_url(
                maximum_layer_order=None, transformation_instance_list=None,
                user=user, viewname='rest_api:trasheddocument-image',
                view_kwargs={'document_id': self.pk}
            )
        else:
            raise AppImageError(error_name=IMAGE_ERROR_NO_VERSION_PAGES)

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_trashed_document_restored,
        target='self',
    )
    def restore(self):
        self.in_trash = False
        # Skip the edit event at .save().
        self._event_ignore = True
        self.save(update_fields=('in_trash',))



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0022_workflow_auto_launch'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workflow',
            name='auto_launch',
            field=models.BooleanField(
                default=True, help_text='Launch workflow when document is '
                'created.', verbose_name='Auto launch'
            ),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_DOCUMENTS_DISPLAY_HEIGHT, DEFAULT_DOCUMENTS_DISPLAY_WIDTH,
    DEFAULT_DOCUMENTS_FAVORITE_COUNT,
    DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND,
    DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_MAXIMUM_SIZE,
    DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND,
    DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_DOCUMENTS_HASH_BLOCK_SIZE, DEFAULT_DOCUMENTS_LIST_THUMBNAIL_WIDTH,
    DEFAULT_DOCUMENTS_PREVIEW_HEIGHT, DEFAULT_DOCUMENTS_PREVIEW_WIDTH,
    DEFAULT_DOCUMENTS_PRINT_HEIGHT, DEFAULT_DOCUMENTS_PRINT_WIDTH,
    DEFAULT_DOCUMENTS_RECENTLY_ACCESSED_COUNT,
    DEFAULT_DOCUMENTS_RECENTLY_CREATED_COUNT, DEFAULT_DOCUMENTS_ROTATION_STEP,
    DEFAULT_DOCUMENTS_THUMBNAIL_HEIGHT, DEFAULT_DOCUMENTS_THUMBNAIL_WIDTH,
    DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_MAXIMUM_SIZE,
    DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND,
    DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_DOCUMENTS_ZOOM_MAX_LEVEL, DEFAULT_DOCUMENTS_ZOOM_MIN_LEVEL,
    DEFAULT_DOCUMENTS_ZOOM_PERCENT_STEP, DEFAULT_LANGUAGE,
    DEFAULT_LANGUAGE_CODES, DEFAULT_STUB_EXPIRATION_INTERVAL
)
from .setting_callbacks import (
    callback_update_document_file_page_image_cache_size,
    callback_update_document_version_page_image_cache_size
)
from .setting_migrations import DocumentsSettingMigration

namespace = SettingNamespace(
    label=_('Documents'), migration_class=DocumentsSettingMigration,
    name='documents', version='0004'
)

setting_display_height = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_DISPLAY_HEIGHT,
    global_name='DOCUMENTS_DISPLAY_HEIGHT'
)
setting_display_width = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_DISPLAY_WIDTH,
    global_name='DOCUMENTS_DISPLAY_WIDTH'
)
setting_document_file_page_image_cache_maximum_size = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_MAXIMUM_SIZE,
    global_name='DOCUMENTS_FILE_PAGE_IMAGE_CACHE_MAXIMUM_SIZE',
    help_text=_(
        'The threshold at which the DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND will start '
        'deleting the oldest document file page image cache files. Specify '
        'the size in bytes.'
    ), post_edit_function=callback_update_document_file_page_image_cache_size
)
setting_document_file_storage_backend = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND,
    global_name='DOCUMENTS_FILE_STORAGE_BACKEND', help_text=_(
        'Path to the Storage subclass to use when storing document '
        'files.'
    )
)
setting_document_file_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND_ARGUMENTS,
    global_name='DOCUMENTS_FILE_STORAGE_BACKEND_ARGUMENTS', help_text=_(
        'Arguments to pass to the DOCUMENT_FILE_STORAGE_BACKEND.'
    )
)
setting_document_file_page_image_cache_storage_backend = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND,
    global_name='DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND', help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'document file page image files.'
    )
)
setting_document_file_page_image_cache_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND.'
    ),
)
setting_favorite_count = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_FAVORITE_COUNT,
    global_name='DOCUMENTS_FAVORITE_COUNT', help_text=_(
        'Maximum number of favorite documents to remember per user.'
    )
)
setting_hash_block_size = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_HASH_BLOCK_SIZE,
    global_name='DOCUMENTS_HASH_BLOCK_SIZE', help_text=_(
        'Size of blocks to use when calculating the document file\'s '
        'checksum. A value of 0 disables the block calculation and the entire '
        'file will be loaded into memory.'
    )
)
setting_language = namespace.add_setting(
    default=DEFAULT_LANGUAGE, global_name='DOCUMENTS_LANGUAGE',
    help_text=_('Default documents language (in ISO639-3 format).')
)
setting_language_codes = namespace.add_setting(
    default=DEFAULT_LANGUAGE_CODES, global_name='DOCUMENTS_LANGUAGE_CODES',
    help_text=_('List of supported document languages. In ISO639-3 format.')
)
setting_document_version_page_image_cache_maximum_size = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_MAXIMUM_SIZE,
    global_name='DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_MAXIMUM_SIZE',
    help_text=_(
        'The threshold at which the DOCUMENT_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND will start '
        'deleting the oldest document version page image cache versions. Specify '
        'the size in bytes.'
    ), post_edit_function=callback_update_document_version_page_image_cache_size
)
setting_document_version_page_image_cache_storage_backend = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND,
    global_name='DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND',
    help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'document version page image versions.'
    )
)
setting_document_version_page_image_cache_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND.'
    ),
)
setting_preview_height = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_PREVIEW_HEIGHT,
    global_name='DOCUMENTS_PREVIEW_HEIGHT'
)
setting_preview_width = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_PREVIEW_WIDTH,
    global_name='DOCUMENTS_PREVIEW_WIDTH'
)
setting_print_height = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_PRINT_HEIGHT,
    global_name='DOCUMENTS_PRINT_HEIGHT'
)
setting_print_width = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_PRINT_WIDTH,
    global_name='DOCUMENTS_PRINT_WIDTH'
)
setting_recently_accessed_document_count = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_RECENTLY_ACCESSED_COUNT,
    global_name='DOCUMENTS_RECENTLY_ACCESSED_COUNT', help_text=_(
        'Maximum number of recently accessed documents (created, edited, '
        'viewed) to remember per user.'
    )
)
setting_recently_created_document_count = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_RECENTLY_CREATED_COUNT,
    global_name='DOCUMENTS_RECENTLY_CREATED_COUNT', help_text=_(
        'Maximum number of recently created documents to show.'
    )
)
setting_rotation_step = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_ROTATION_STEP,
    global_name='DOCUMENTS_ROTATION_STEP', help_text=_(
        'Amount in degrees to rotate a document page per user interaction.'
    )
)
setting_stub_expiration_interval = namespace.add_setting(
    default=DEFAULT_STUB_EXPIRATION_INTERVAL,
    global_name='DOCUMENTS_STUB_EXPIRATION_INTERVAL', help_text=_(
        'Time after which a document stub will be considered invalid and '
        'deleted.'
    )
)
setting_thumbnail_height = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_THUMBNAIL_HEIGHT,
    global_name='DOCUMENTS_THUMBNAIL_HEIGHT', help_text=_(
        'Height in pixels of the document thumbnail image.'
    )
)
setting_thumbnail_width = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_THUMBNAIL_WIDTH,
    global_name='DOCUMENTS_THUMBNAIL_WIDTH', help_text=(
        'Width in pixels of the document thumbnail image.'
    )
)
setting_thumbnail_list_width = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_LIST_THUMBNAIL_WIDTH,
    global_name='DOCUMENTS_LIST_THUMBNAIL_WIDTH', help_text=(
        'Width in pixels of the document thumbnail image when shown in list '
        'view mode.'
    )
)
setting_zoom_max_level = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_ZOOM_MAX_LEVEL,
    global_name='DOCUMENTS_ZOOM_MAX_LEVEL', help_text=_(
        'Maximum amount in percent (%) to allow user to zoom in a document '
        'page interactively.'
    )
)
setting_zoom_min_level = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_ZOOM_MIN_LEVEL,
    global_name='DOCUMENTS_ZOOM_MIN_LEVEL', help_text=_(
        'Minimum amount in percent (%) to allow user to zoom out a document '
        'page interactively.'
    )
)
setting_zoom_percent_step = namespace.add_setting(
    default=DEFAULT_DOCUMENTS_ZOOM_PERCENT_STEP,
    global_name='DOCUMENTS_ZOOM_PERCENT_STEP', help_text=_(
        'Amount in percent zoom in or out a document page per user '
        'interaction.'
    )
)



import hashlib
import io
import logging
import uuid

from PIL import Image
from furl import furl

from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Max
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext, ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.common.validators import (
    YAMLValidator, validate_internal_name
)
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event
from mayan.apps.file_caching.models import CachePartitionFile
from mayan.apps.storage.classes import DefinedStorageLazy

from .classes import Layer
from .events import event_asset_created, event_asset_edited
from .literals import STORAGE_NAME_ASSETS, STORAGE_NAME_ASSETS_CACHE
from .managers import LayerTransformationManager, ObjectLayerManager
from .transformations import BaseTransformation

logger = logging.getLogger(name=__name__)


def upload_to(instance, filename):
    return 'converter-asset-{}'.format(uuid.uuid4().hex)


class Asset(ExtraDataModelMixin, models.Model):
    """
    This model keeps track of files that will be available for use with
    transformations.
    """
    label = models.CharField(
        max_length=96, unique=True, verbose_name=_('Label')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used when referencing this asset. '
            'Can only contain letters, numbers, and underscores.'
        ), max_length=255, unique=True, validators=[validate_internal_name],
        verbose_name=_('Internal name')
    )
    file = models.FileField(
        storage=DefinedStorageLazy(name=STORAGE_NAME_ASSETS),
        upload_to=upload_to, verbose_name=_('File')
    )

    class Meta:
        ordering = ('label',)
        verbose_name = _('Asset')
        verbose_name_plural = _('Assets')

    def __str__(self):
        return self.label

    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_ASSETS_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        name = self.file.name
        self.file.close()
        self.file.storage.delete(name=name)
        return super().delete(*args, **kwargs)

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # The parameters 'maximum_layer_order',
        # `transformation_instance_list`, `user` are not used, but added
        # to retain interface compatibility.
        cache_filename = '{}'.format(self.get_hash())

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug(
                'asset cache file "%s" not found', cache_filename
            )

            image = self.get_image()
            with io.BytesIO() as image_buffer:
                image.save(image_buffer, format='PNG')

                with self.cache_partition.create_file(filename=cache_filename) as file_object:
                    file_object.write(image_buffer.getvalue())
        else:
            logger.debug(
                'asset cache file "%s" found', cache_filename
            )

        return cache_filename

    def get_absolute_url(self):
        return reverse(
            viewname='converter:asset_detail', kwargs={
                'asset_id': self.pk
            }
        )

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:asset-image',
            kwargs={'asset_id': self.pk}
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_hash(self):
        with self.open() as file_object:
            return hashlib.sha256(file_object.read()).hexdigest()

    def get_image(self):
        with self.open() as file_object:
            image = Image.open(fp=file_object)
            image.load()

            if image.mode != 'RGBA':
                image.putalpha(alpha=255)

        return image

    def open(self):
        name = self.file.name
        self.file.close()
        return self.file.storage.open(name=name)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'event': event_asset_created,
            'target': 'self',
        },
        edited={
            'event': event_asset_edited,
            'target': 'self',
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)


class StoredLayer(models.Model):
    name = models.CharField(
        max_length=64, unique=True, verbose_name=_('Name')
    )
    order = models.PositiveIntegerField(
        db_index=True, unique=True, verbose_name=_('Order')
    )

    class Meta:
        ordering = ('order',)
        verbose_name = _('Stored layer')
        verbose_name_plural = _('Stored layers')

    def __str__(self):
        return self.name

    def get_layer(self):
        return Layer.get(name=self.name)


class ObjectLayer(models.Model):
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey(
        ct_field='content_type', fk_field='object_id'
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))
    stored_layer = models.ForeignKey(
        on_delete=models.CASCADE, related_name='object_layers', to=StoredLayer,
        verbose_name=_('Stored layer')
    )

    objects = ObjectLayerManager()

    class Meta:
        ordering = ('stored_layer__order',)
        unique_together = ('content_type', 'object_id', 'stored_layer')
        verbose_name = _('Object layer')
        verbose_name_plural = _('Object layers')

    def get_next_order(self):
        last_order = self.transformations.aggregate(
            Max('order')
        )['order__max']

        if last_order is not None:
            return last_order + 1
        else:
            return 0


class LayerTransformation(models.Model):
    """
    Model that stores the transformation and transformation arguments
    for a given object
    Fields:
    * order - Order of a Transformation - In case there are multiple
    transformations for an object, this field list the order at which
    they will be execute.
    * arguments - Arguments of a Transformation - An optional field to hold a
    transformation argument. Example: if a page is rotated with the Rotation
    transformation, this field will show by how many degrees it was rotated.
    """
    object_layer = models.ForeignKey(
        on_delete=models.CASCADE, related_name='transformations',
        to=ObjectLayer, verbose_name=_('Object layer')
    )
    order = models.PositiveIntegerField(
        blank=True, db_index=True, default=0, help_text=_(
            'Order in which the transformations will be executed. If left '
            'unchanged, an automatic order value will be assigned.'
        ), verbose_name=_('Order')
    )
    name = models.CharField(max_length=128, verbose_name=_('Name'))
    arguments = models.TextField(
        blank=True, help_text=_(
            'Enter the arguments for the transformation as a YAML '
            'dictionary. ie: {"degrees": 180}'
        ), validators=[YAMLValidator()], verbose_name=_('Arguments')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))

    objects = LayerTransformationManager()

    class Meta:
        ordering = ('object_layer__stored_layer__order', 'order',)
        unique_together = ('object_layer', 'order')
        verbose_name = _('Layer transformation')
        verbose_name_plural = _('Layer transformations')

    def __str__(self):
        try:
            return str(BaseTransformation.get(name=self.name))
        except KeyError:
            return ugettext('Unknown transformation class')

    def get_transformation_class(self):
        return BaseTransformation.get(name=self.name)

    def save(self, *args, **kwargs):
        if not self.order:
            self.order = self.object_layer.get_next_order()
        super().save(*args, **kwargs)



from django.test import override_settings

from mayan.apps.testing.tests.base import GenericViewTestCase
from mayan.apps.views.http import URL


TEST_INSTALLATION_SCHEME = 'https'
TEST_INSTALLATION_NETLOC = 'example.com:9999'

TEST_INSTALLATION_URL = URL(
    netloc=TEST_INSTALLATION_NETLOC, scheme=TEST_INSTALLATION_SCHEME
).to_string()


class RequestTestCase(GenericViewTestCase):
    @override_settings(ORGANIZATIONS_INSTALLATION_URL=TEST_INSTALLATION_URL)
    def test_installation_url_override(self):
        self.add_test_view()
        context = self.get_test_view()

        absolute_url = context['request'].build_absolute_uri()

        url = URL(url=absolute_url)

        self.assertEqual(url._split_result.netloc, TEST_INSTALLATION_NETLOC)
        self.assertEqual(url._split_result.scheme, TEST_INSTALLATION_SCHEME)

    def test_original_absolute_uri(self):
        self.add_test_view()
        context = self.get_test_view()

        absolute_url = context['request'].build_absolute_uri()

        url = URL(url=absolute_url)

        self.assertNotEqual(url._split_result.netloc, TEST_INSTALLATION_NETLOC)
        self.assertNotEqual(url._split_result.scheme, TEST_INSTALLATION_SCHEME)



from django.contrib.auth.backends import BaseBackend, ModelBackend

from mayan.apps.authentication.django_authentication_backends import (
    DjangoAuthenticationBackendModelEmail,
    DjangoAuthenticationBackendWrapperMultiFactor
)


class DjangoAuthenticationBackendOTP(BaseBackend):
    def authenticate(self, request, otp_token, user):
        if user:
            if user.otp_data.is_enabled():
                if user.otp_data.verify_token(token=otp_token):
                    return user
                else:
                    return
            else:
                return user
        else:
            return


class DjangoAuthenticationBackendEmailMultiFactorOTP(
    DjangoAuthenticationBackendWrapperMultiFactor
):
    factors = (
        {
            'default': True,
            'class': DjangoAuthenticationBackendModelEmail,
            'name': 'email_password'
        },
        {
            'class': DjangoAuthenticationBackendOTP,
            'name': 'otp_token'
        }
    )


class DjangoAuthenticationBackendUsernameMultiFactorOTP(
    DjangoAuthenticationBackendWrapperMultiFactor
):
    factors = (
        {
            'default': True,
            'class': ModelBackend,
            'name': 'username_password'
        },
        {
            'class': DjangoAuthenticationBackendOTP,
            'name': 'otp_token'
        }
    )



from django.apps import apps
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.dashboards.classes import DashboardWidgetNumeric

from .icons import icon_role_list
from .permissions import permission_role_view


class DashboardWidgetRoleTotal(DashboardWidgetNumeric):
    icon = icon_role_list
    label = _('Total roles')
    link = reverse_lazy(viewname='permissions:role_list')

    def get_count(self):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        Role = apps.get_model(
            app_label='permissions', model_name='Role'
        )

        return AccessControlList.objects.restrict_queryset(
            permission=permission_role_view, user=self.request.user,
            queryset=Role.objects.all()
        ).count()



from django.db import migrations


def code_delete_all_error_log_partitions(apps, schema_editor):
    ErrorLogPartition = apps.get_model(
        app_label='logging', model_name='ErrorLogPartition'
    )

    ErrorLogPartition.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('logging', '0004_auto_20220118_2327'),
    ]

    operations = [
        migrations.RunPython(
            code=code_delete_all_error_log_partitions,
            reverse_code=migrations.RunPython.noop
        )
    ]



from mayan.apps.common.compatibility import Iterable
from mayan.apps.common.utils import (
    ResolverPipelineModelAttribute, flatten_list
)

from .classes import SearchBackend
from .tasks import (
    task_deindex_instance, task_index_instance,
    task_index_related_instance_m2m
)


def handler_deindex_instance(sender, **kwargs):
    instance = kwargs['instance']

    task_deindex_instance.apply_async(
        kwargs={
            'app_label': instance._meta.app_label,
            'model_name': instance._meta.model_name,
            'object_id': instance.pk
        }
    )


def handler_factory_index_related_instance_delete(reverse_field_path):
    def handler_index_by_related_to_delete_instance(sender, **kwargs):
        related_instance = kwargs['instance']

        result = ResolverPipelineModelAttribute.resolve(
            attribute=reverse_field_path, obj=related_instance
        )

        entries = flatten_list(value=result)

        def call_task(instance):
            task_index_instance.apply_async(
                kwargs={
                    'app_label': instance._meta.app_label,
                    'model_name': instance._meta.model_name,
                    'object_id': instance.pk,
                    'exclude_app_label': related_instance._meta.app_label,
                    'exclude_model_name': related_instance._meta.model_name,
                    'exclude_kwargs': {'id': related_instance.pk}
                }
            )

        if isinstance(entries, Iterable):
            for instance in entries:
                call_task(instance=instance)
        else:
            call_task(instance=result)

    return handler_index_by_related_to_delete_instance


def handler_factory_index_related_instance_save(reverse_field_path):
    def handler_index_by_related_instance(sender, **kwargs):
        related_instance = kwargs['instance']

        result = ResolverPipelineModelAttribute.resolve(
            attribute=reverse_field_path, obj=related_instance
        )

        entries = flatten_list(value=result)

        def call_task(instance):
            task_index_instance.apply_async(
                kwargs={
                    'app_label': instance._meta.app_label,
                    'model_name': instance._meta.model_name,
                    'object_id': instance.pk
                }
            )

        if isinstance(entries, Iterable):
            for instance in entries:
                call_task(instance=instance)
        else:
            call_task(instance=result)

    return handler_index_by_related_instance


def handler_factory_index_related_instance_m2m(data):
    # Serialize search model field paths.
    serialized_search_model_related_paths = {}

    for key, value in data.items():
        serialized_search_model_related_paths[
            '{}.{}'.format(key._meta.app_label, key._meta.model_name)
        ] = tuple(value)

    def handler_index_related_instance_m2m(sender, **kwargs):
        action = kwargs.get('action')
        instance = kwargs['instance']
        model = kwargs.get('model')

        kwargs = {
            'action': action,
            'instance_app_label': instance._meta.app_label,
            'instance_model_name': instance._meta.model_name,
            'instance_object_id': instance.pk,
            'model_app_label': model._meta.app_label,
            'model_model_name': model._meta.model_name,
            'pk_set': tuple(kwargs['pk_set']),
            'serialized_search_model_related_paths': serialized_search_model_related_paths
        }

        task_index_related_instance_m2m.apply_async(
            kwargs=kwargs
        )

    return handler_index_related_instance_m2m


def handler_index_instance(sender, **kwargs):
    instance = kwargs['instance']

    task_index_instance.apply_async(
        kwargs={
            'app_label': instance._meta.app_label,
            'model_name': instance._meta.model_name,
            'object_id': instance.pk
        }
    )


def handler_search_backend_initialize(sender, **kwargs):
    backend = SearchBackend.get_instance()

    backend.initialize()


def handler_search_backend_upgrade(sender, **kwargs):
    backend = SearchBackend.get_instance()

    backend.upgrade()



import json
import shutil

from mayan.apps.documents.tests.literals import TEST_SMALL_DOCUMENT_PATH
from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ...source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from ...source_backends.staging_folder_backends import SourceBackendStagingFolder

from ..literals import (
    TEST_STAGING_PREVIEW_HEIGHT, TEST_STAGING_PREVIEW_WIDTH
)

from .base_mixins import SourceTestMixin


class StagingFolderActionAPIViewTestMixin:
    def _request_test_staging_folder_file_delete_action_api_view(self):
        return self.post(
            viewname='rest_api:source-action', kwargs={
                'action_name': 'file_delete', 'source_id': self.test_source.pk
            }, data={
                'arguments': json.dumps(
                    obj={
                        'encoded_filename': self.test_staging_folder_file.encoded_filename
                    }
                )
            }
        )

    def _request_test_staging_folder_file_image_action_api_view(self):
        return self.get(
            viewname='rest_api:source-action', kwargs={
                'action_name': 'file_image', 'source_id': self.test_source.pk
            }, query={'encoded_filename': self.test_staging_folder_file.encoded_filename}
        )

    def _request_test_staging_folder_file_list_action_api_view(self):
        return self.get(
            viewname='rest_api:source-action', kwargs={
                'action_name': 'file_list', 'source_id': self.test_source.pk
            }
        )

    def _request_test_staging_folder_file_upload_action_api_view(self):
        return self.post(
            viewname='rest_api:source-action', kwargs={
                'action_name': 'file_upload', 'source_id': self.test_source.pk
            }, data={
                'arguments': json.dumps(
                    obj={
                        'document_type_id': self.test_document_type.pk,
                        'encoded_filename': self.test_staging_folder_file.encoded_filename
                    }
                )
            }
        )


class StagingFolderTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_staging_folder'

    def setUp(self):
        self._temporary_folders = []
        super().setUp()
        self.test_staging_folder_files = []

    def tearDown(self):
        for temporary_folders in self._temporary_folders:
            fs_cleanup(filename=temporary_folders)

        super().tearDown()

    def _create_test_staging_folder(self, extra_data=None):
        temporary_folder = mkdtemp()
        self._temporary_folders.append(temporary_folder)
        backend_data = {
            'folder_path': temporary_folder,
            'preview_width': TEST_STAGING_PREVIEW_WIDTH,
            'preview_height': TEST_STAGING_PREVIEW_HEIGHT,
            'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER
        }

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendStagingFolder.get_class_path(),
            backend_data=backend_data
        )

    def _copy_test_staging_folder_document(self):
        shutil.copy(
            src=TEST_SMALL_DOCUMENT_PATH,
            dst=self.test_source.get_backend_data()['folder_path']
        )
        self.test_staging_folder_file = list(
            self.test_source.get_backend_instance().get_files()
        )[0]
        self.test_staging_folder_files.append(self.test_staging_folder_file)


class StagingFolderViewTestMixin:
    def _request_staging_folder_action_file_delete_view_via_get(self):
        return self.get(
            viewname='sources:source_action', kwargs={
                'source_id': self.test_source.pk,
                'action_name': 'file_delete'
            }, query={
                'encoded_filename': self.test_staging_folder_file.encoded_filename
            }
        )

    def _request_staging_folder_action_file_delete_view_via_post(self):
        return self.post(
            viewname='sources:source_action', kwargs={
                'source_id': self.test_source.pk,
                'action_name': 'file_delete'
            }, query={
                'encoded_filename': self.test_staging_folder_file.encoded_filename
            }
        )



import os

from django.conf import settings

DEFAULT_ERROR_LOG_PARTITION_ENTRY_LIMIT = 3
DEFAULT_LOGGING_ENABLE = True
DEFAULT_LOGGING_HANDLERS = ('console',)
DEFAULT_LOGGING_LEVEL = 'ERROR'
DEFAULT_LOGGING_LOG_FILE_PATH = os.path.join(settings.MEDIA_ROOT, 'error.log')

FORMATTER_PALETTE = {
    'CRITICAL': {'fg': 'red', 'opts': ('bold', 'blink', 'reverse')},
    'DEBUG': {'fg': 'cyan'},
    'ERROR': {'fg': 'red', 'opts': ('bold',)},
    'INFO': {'fg': 'white'},
    'SUCCESS': {'fg': 'green'},
    'WARNING': {'fg': 'yellow', 'opts': ('bold', 'underscore')},
}

LOGGING_HANDLER_OPTIONS = ('console', 'logfile')



from rest_framework import status

from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_type_view
)
from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_web_link_created, event_web_link_edited
from ..models import WebLink
from ..permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_view
)

from .literals import TEST_WEB_LINK_LABEL_EDITED, TEST_WEB_LINK_LABEL
from .mixins import (
    WebLinkDocumentTypeAPIViewMixin, WebLinkAPIViewTestMixin,
    WebLinkTestMixin
)


class WebLinkAPIViewTestCase(
    WebLinkTestMixin, WebLinkAPIViewTestMixin, BaseAPITestCase
):
    def test_web_link_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_web_link_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(WebLink.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_web_link_create)

        self._clear_events()

        response = self._request_test_web_link_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        web_link = WebLink.objects.first()
        self.assertEqual(response.data['id'], web_link.pk)
        self.assertEqual(response.data['label'], TEST_WEB_LINK_LABEL)

        self.assertEqual(WebLink.objects.count(), 1)
        self.assertEqual(web_link.label, TEST_WEB_LINK_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_web_link)
        self.assertEqual(events[0].verb, event_web_link_created.id)

    def test_web_link_delete_api_view_no_permission(self):
        self._create_test_web_link()

        self._clear_events()

        response = self._request_test_web_link_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(WebLink.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_delete_api_view_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_delete
        )

        self._clear_events()

        response = self._request_test_web_link_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(WebLink.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_detail_api_view_no_permission(self):
        self._create_test_web_link()

        self._clear_events()

        response = self._request_test_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertFalse('label' in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_detail_api_view_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_view
        )

        self._clear_events()

        response = self._request_test_web_link_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['label'], TEST_WEB_LINK_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_edit_api_view_via_patch_no_permission(self):
        self._create_test_web_link()

        self._clear_events()

        response = self._request_test_web_link_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()
        self.assertEqual(self.test_web_link.label, TEST_WEB_LINK_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_edit_api_view_via_patch_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()
        self.assertEqual(
            self.test_web_link.label, TEST_WEB_LINK_LABEL_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)

    def test_web_link_edit_api_view_via_put_no_permission(self):
        self._create_test_web_link()

        self._clear_events()

        response = self._request_test_web_link_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()
        self.assertEqual(self.test_web_link.label, TEST_WEB_LINK_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_edit_api_view_via_put_with_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        self._clear_events()

        response = self._request_test_web_link_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()
        self.assertEqual(
            self.test_web_link.label, TEST_WEB_LINK_LABEL_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)


class WebLinkDocumentTypeAPIViewTestCase(
    DocumentTestMixin, WebLinkDocumentTypeAPIViewMixin, WebLinkTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def test_web_link_document_type_add_api_view_no_permission(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_api_view_with_document_type_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_api_view_with_web_link_access(self):
        self._create_test_web_link()

        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_add_api_view_with_full_access(self):
        self._create_test_web_link()
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)

    def test_web_link_permission_list_api_view_no_permission(self):
        self._create_test_web_link()
        self.test_web_link.document_types.add(self.test_document_type)

        self._clear_events()

        response = self._request_test_web_link_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_permission_list_api_view_with_document_type_access(self):
        self._create_test_web_link()
        self.test_web_link.document_types.add(self.test_document_type)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_permission_list_api_view_with_web_link_access(self):
        self._create_test_web_link()
        self.test_web_link.document_types.add(self.test_document_type)

        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_view
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_permission_list_api_view_with_full_access(self):
        self._create_test_web_link()
        self.test_web_link.document_types.add(self.test_document_type)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_view
        )

        self._clear_events()

        response = self._request_test_web_link_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self.test_document_type.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_api_view_no_permission(self):
        self._create_test_web_link(add_test_document_type=True)
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_api_view_with_document_type_access(self):
        self._create_test_web_link(add_test_document_type=True)
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_api_view_with_web_link_access(self):
        self._create_test_web_link(add_test_document_type=True)

        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_link_document_type_remove_api_view_with_full_access(self):
        self._create_test_web_link(add_test_document_type=True)
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_web_link, permission=permission_web_link_edit
        )

        test_web_link_document_types_count = self.test_web_link.document_types.count()

        self._clear_events()

        response = self._request_test_web_link_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_web_link.refresh_from_db()

        self.assertEqual(
            self.test_web_link.document_types.count(),
            test_web_link_document_types_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_web_link)
        self.assertEqual(events[0].verb, event_web_link_edited.id)



import logging

from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import OperationalError
from django.utils.module_loading import import_string

from mayan.celery import app

from .literals import (
    UPDATE_PAGE_COUNT_RETRY_DELAY, UPLOAD_NEW_VERSION_RETRY_DELAY
)

logger = logging.getLogger(name=__name__)


# Document file

@app.task(
    bind=True, default_retry_delay=UPDATE_PAGE_COUNT_RETRY_DELAY,
    ignore_result=True
)
def task_document_file_page_count_update(
    self, document_file_id, user_id=None
):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_file = DocumentFile.objects.get(pk=document_file_id)
    try:
        document_file.page_count_update(user=user)
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to update page count for '
            'document file: %s; %s. Retrying.', document_file,
            exception
        )
        raise self.retry(exc=exception)


@app.task(
    bind=True, default_retry_delay=UPLOAD_NEW_VERSION_RETRY_DELAY,
    ignore_result=True
)
def task_document_file_upload(
    self, document_id, shared_uploaded_file_id, user_id=None, action=None,
    comment=None, expand=False, filename=None
):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )

    SharedUploadedFile = apps.get_model(
        app_label='storage', model_name='SharedUploadedFile'
    )

    try:
        document = Document.objects.get(pk=document_id)
        shared_uploaded_file = SharedUploadedFile.objects.get(
            pk=shared_uploaded_file_id
        )
        if user_id:
            user = get_user_model().objects.get(pk=user_id)
        else:
            user = None

    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to retrieve shared data for '
            'new document file for document ID: %s; %s. Retrying.', document_id,
            exception
        )
        raise self.retry(exc=exception)

    with shared_uploaded_file.open() as file_object:
        try:
            document.file_new(
                action=action, comment=comment, expand=expand,
                file_object=file_object,
                filename=filename or shared_uploaded_file.filename,
                _user=user
            )
        except Warning as warning:
            # New document file are blocked
            logger.info(
                'Warning during attempt to create new document file for '
                'document: %s; %s', document, warning
            )
            shared_uploaded_file.delete()
        except OperationalError as exception:
            logger.warning(
                'Operational error during attempt to create new document '
                'file for document: %s; %s. Retrying.', document, exception
            )
            raise self.retry(exc=exception)
        except Exception as exception:
            # This except and else block emulate a finally:
            logger.error(
                'Unexpected error during attempt to create new document '
                'file for document: %s; %s', document, exception,
                exc_info=True
            )
            try:
                shared_uploaded_file.delete()
            except OperationalError as exception:
                logger.warning(
                    'Operational error during attempt to delete shared '
                    'file: %s; %s.', shared_uploaded_file, exception
                )
        else:
            try:
                shared_uploaded_file.delete()
            except OperationalError as exception:
                logger.warning(
                    'Operational error during attempt to delete shared '
                    'file: %s; %s.', shared_uploaded_file, exception
                )


# Document

@app.task(ignore_result=True)
def task_document_stubs_delete():
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )

    logger.info(msg='Executing')
    Document.objects.delete_stubs()
    logger.info(msg='Finished')


@app.task(ignore_results=True)
def task_document_upload(
    document_type_id, shared_uploaded_file_id, callback_dotted_path=None,
    callback_function=None, callback_kwargs=None, description=None,
    label=None, language=None, user_id=None
):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    SharedUploadedFile = apps.get_model(
        app_label='storage', model_name='SharedUploadedFile'
    )

    document_type = DocumentType.objects.get(pk=document_type_id)
    shared_uploaded_file = SharedUploadedFile.objects.get(
        pk=shared_uploaded_file_id
    )

    if user_id:
        user = get_user_model().objects.get(pk=user_id)
    else:
        user = None

    document = None
    try:
        with shared_uploaded_file.open() as file_object:
            document, document_file = document_type.new_document(
                file_object=file_object,
                label=label or shared_uploaded_file.filename,
                description=description, language=language,
                _user=user
            )
    except Exception as exception:
        logger.critical(
            'Unexpected exception while trying to create new document '
            '"%s"; %s',
            label or file_object.name, exception
        )
        if document:
            document.delete(to_trash=False)
        raise
    else:
        shared_uploaded_file.delete()

        if user:
            document.add_as_recent_document_for_user(user=user)

    if callback_dotted_path:
        callback = import_string(dotted_path=callback_dotted_path)
        callback_kwargs = callback_kwargs or {}
        function = getattr(callback, callback_function)
        function(
            document_file=document_file, **callback_kwargs
        )


# Document type

@app.task(ignore_result=True)
def task_document_type_trashed_document_delete_periods_check():
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    DocumentType.objects.check_delete_periods()


@app.task(ignore_result=True)
def task_document_type_document_trash_periods_check():
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )

    DocumentType.objects.check_trash_periods()


# Document version

@app.task(ignore_result=True)
def task_document_version_page_list_append(document_version_id, user_id=None):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )
    document_version.pages_append_all(_user=user)


@app.task(ignore_result=True)
def task_document_version_page_list_reset(document_version_id, user_id=None):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )
    document_version.pages_reset(_user=user)


@app.task(ignore_result=True)
def task_document_version_delete(document_version_id, user_id=None):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    document_version._event_actor = user
    document_version.delete()


@app.task(ignore_result=True)
def task_document_version_export(
    document_version_id, organization_installation_url=None, user_id=None
):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    document_version.export_to_download_file(
        organization_installation_url=organization_installation_url, user=user
    )


# Trash can

@app.task(ignore_result=True)
def task_trash_can_empty(user_id=None):
    TrashedDocument = apps.get_model(
        app_label='documents', model_name='TrashedDocument'
    )

    for trashed_document in TrashedDocument.objects.all():
        task_trashed_document_delete.apply_async(
            kwargs={
                'trashed_document_id': trashed_document.pk,
                'user_id': user_id
            }
        )


# Trashed document

@app.task(ignore_result=True)
def task_trashed_document_delete(trashed_document_id, user_id=None):
    TrashedDocument = apps.get_model(
        app_label='documents', model_name='TrashedDocument'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    logger.debug(msg='Executing')
    trashed_document = TrashedDocument.objects.get(pk=trashed_document_id)
    trashed_document._event_actor = user
    trashed_document.delete()
    logger.debug(msg='Finished')



from mayan.apps.appearance.classes import Icon

icon_api = Icon(driver_name='fontawesome', symbol='plug')
icon_api_documentation = Icon(driver_name='fontawesome', symbol='book')
icon_api_documentation_redoc = Icon(driver_name='fontawesome', symbol='book')
